;;; -*- Package: C; Log: C.Log -*-
;;;
;;; **********************************************************************
;;; This code was written as part of the Spice Lisp project at
;;; Carnegie-Mellon University, and has been placed in the public domain.
;;; If you want to use this code or any part of Spice Lisp, please contact
;;; Scott Fahlman (FAHLMAN@CMUC). 
;;; **********************************************************************
;;;
;;;    This file implements the IR1 finalize phase, which checks for various
;;; semantic errors.
;;;
;;; Written by Rob MacLachlan
;;;
(in-package 'c)


;;; IR1-Finalize  --  Interface
;;;
;;; We do a number of things:
;;;  2] Find any unknown free functions.
;;;  3] Accumulate approximate function type info for unknown functions.
;;;  4] Update the global type for functions newly defined.
;;;  5] Emit any delayed notes about failed optimizations.
;;;
(proclaim '(function ir1-finalize () void))
(defun ir1-finalize ()
  (maphash #'check-free-function *free-functions*)
  (maphash #'note-failed-optimization *failed-optimizations*))


;;; Note-Failed-Optimization  --  Internal
;;;
;;;    Give the user grief about optimizations that we weren't able to do.  It
;;; is assumed that they want to hear, or there wouldn't be any entries in the
;;; table.  If the node has been deleted or is no longer a known call, then do
;;; nothing; some other optimization must have gotten to it.
;;;
(proclaim '(function note-failed-optimization (combination (or list ctype))
		     void))
(defun note-failed-optimization (node what)
  (unless (or (node-deleted node)
	      (not (function-info-p (combination-kind node))))
    (let ((*compiler-error-context* node))
      (cond ((listp what)
	     (compiler-note "Unable to optimize because:~%~6T~?"
			    (first what) (rest what)))
	    (t
	     (collect ((messages))
	       (flet ((frob (string &rest stuff)
			(messages string)
			(messages stuff)))
		 (valid-function-use node what
				     :warning-function #'frob
				     :error-function #'frob))
	       
	       (compiler-note "Unable to optimize due to type uncertainty:~@
	                       ~{~6T~?~^~&~}"
			      (messages))))))))

	  
;;; Check-Free-Function  --  Internal
;;;
;;;    If the entry is a functional, then we update the global environment
;;; according to the new definition, checking for inconsistency.  If the entry
;;; is an unknown global function, then we add the uses into the function's
;;; approximate type.
;;;
(proclaim '(function check-free-function (t leaf) void))
(defun check-free-function (name leaf)
  (etypecase leaf
    (functional
     (let* ((where (info function where-from name))
;	    (type (info function type name))
	    (dtype (leaf-type leaf))
	    (node (lambda-bind (main-entry leaf)))
	    (*compiler-error-context* node))
       (note-name-defined name :function)

       (ecase where
	 (:assumed
	  (let ((approx-type (info function assumed-type name)))
	    (when approx-type
	      (valid-approximate-type approx-type dtype))))
	 ((:declared :defined)
;	  ...
	  ))

       (when (and (eq (function-type-returns dtype) *empty-type*)
		  (policy node (>= safety brevity)))
	 (compiler-note "Function does not return."))

       (setf (info function kind name) :function)
       (setf (info function where-from name) :defined)
       (setf (info function type name) dtype)
       (clear-info function assumed-type name)))
    (global-var)))
